/**
 * John David Ratliff
 * jdratlif@cs.indiana.edu
 * CSCI C311 - Prof. Amy Sabry
 * Monday Lab - 2:30 - 3:20
 */

package scheme.interp;

import scheme.util.*;
import scheme.absyn.*;

public class Eval_Exp implements ExpVisitor {
    private Eval_ExpList eval_explist;
    private Eval_Constant eval_const;
    private Eval_Datum eval_datum;
    private Env env;
    private Continuation cont;

    public Env getEnv() { return env; }

    public void setEnv (Env newenv) { env = newenv; }

    public Continuation getCont () { return cont; }

    public void setCont (Continuation newcont) { cont = newcont; }

    public Eval_Exp () {
        this.eval_const = new Eval_Constant();
        this.eval_datum = new Eval_Datum(eval_const);
        this.eval_explist = new Eval_ExpList(this);
        this.env = new EmptyEnv(); 
        this.cont = new Continuation () {
                public void invoke (SchemeValue v) throws Done {
                    throw new Done(v);
                }
            };
    }

    public void visitConstantE (ConstantE e) throws VisitE {
        e.get().accept(eval_const);
        cont.invoke(eval_const.getValue());
        throw new Error(); // can never get here
    }

    public void visitQuoteE (QuoteE e) throws VisitE {
        e.get().accept(eval_datum);
        cont.invoke(eval_datum.getValue());
        throw new Error(); // can never get here
    }

    public void visitVarE (VarE e) throws VisitE {
        cont.invoke(env.find(e.getName()));
        throw new Error(); // can never get here
    }

    public void visitIfE (IfE e) throws VisitE {
        final Exp thenExp = e.getThen();
        final Exp elseExp = e.getElse();
        final Continuation savedCont = cont;
        cont = new Continuation () {
                public void invoke (SchemeValue testValue) throws VisitE {
                    cont = savedCont;
                    if (testValue.isTrue()) 
                        thenExp.accept(Eval_Exp.this);
                    else 
                        elseExp.accept(Eval_Exp.this);
                    throw new Error(); // can never get here
                }
            };
        Exp testExp = e.getTest();
        testExp.accept(this);
        throw new Error(); // can never get here
    }

    public void visitLambdaE (LambdaE e) throws VisitE {
        // set the value of the continuation as a Closure
        cont.invoke(new ClosureV(e.getIds(), e.getBody(), env));
        throw new Error(); // can never get here
    }

    public void visitSetE (SetE e) throws VisitE {
        final Continuation saved = cont;
        final String x = e.getId();
        
        // setup the continuation
        cont = new Continuation() {
            public void invoke(SchemeValue v) throws VisitE {
                cont = saved;
                env.set(x,v);
                cont.invoke(new VoidV());
            }
        };
        
        // accept the expression
        e.getExp().accept(this);
        throw new Error(); // can never get here...
    }

    public void visitPrimappE (PrimappE e) throws VisitE {
        final Continuation savedCont = cont;
        final byte op = e.getOp();
        
        eval_explist.setContL(new ContinuationL() {
            public void invoke(List vs) throws VisitE {
                cont = savedCont;
                applyPrim(op,vs);
                throw new Error(); // can never get here...
            }
        });
        
        e.getArgs().accept(eval_explist);
        throw new Error(); // can never get here...
    }

    public void visitAppE (AppE e) throws VisitE {
        final Continuation saved = cont;
        final ExpList args = e.getArgs();
        
        cont = new Continuation() {
            public void invoke(final SchemeValue lambda) throws VisitE {
                cont = saved;
                
                eval_explist.setContL(new ContinuationL() {
                    public void invoke(List vs) throws VisitE {
                        cont = saved;
                        lambda.apply(vs,Eval_Exp.this);
                    }
                });
                
                args.accept(eval_explist);
                throw new Error(); // can never get here...
            }
        };
        
        e.getFun().accept(this);
        throw new Error(); // can never get here...
    } 

    public void visitBeginE (BeginE e) throws VisitE {
        final Continuation saved = cont;
        
        eval_explist.setContL(new ContinuationL() {
            public void invoke(List vs) throws VisitE {
                cont = saved;
                cont.invoke((SchemeValue)vs.getLast());
                throw new Error(); // can never get here
            }
        });
        
        e.getExps().accept(eval_explist);
        throw new Error(); // can never get here...
    }

    public void visitLetE (LetE e) throws VisitE {
        final Continuation saved = cont;
        ExpList exps = new EmptyExpL();
        List decls = e.getDecls();
        List vars = new EmptyL();
        
        while (!(decls instanceof EmptyL)) {
            Decl d = (Decl)decls.hd();
            vars = vars.snoc(d.getVar());
            exps = exps.snoc(d.getExp());
            decls = decls.tl();
        }
        
        final List fVars = vars;
        final Exp body = e.getBody();
        
        eval_explist.setContL(new ContinuationL() {
            public void invoke(List vs) throws VisitE {
                cont = saved;
                
                env = env.addAll(fVars,vs);
                body.accept(Eval_Exp.this);
                throw new Error(); // can never get here...
            }
        });
        
        exps.accept(eval_explist);
        throw new Error(); // can never get here...
    }

    public void visitLetrecE (LetrecE e) throws VisitE {
        final Continuation saved = cont;
        ExpList exps = new EmptyExpL();
        List decls = e.getDecls();
        List vars = new EmptyL();
        
        while (!(decls instanceof EmptyL)) {
            Decl d = (Decl)decls.hd();
            vars = vars.snoc(d.getVar());
            exps = exps.snoc(d.getExp());
            decls = decls.tl();
        }
        
        final List fVars = vars;
        final Exp body = e.getBody();
        
        env = env.addAllVoid(vars);
        
        eval_explist.setContL(new ContinuationL() {
            public void invoke(List vs) throws VisitE {
                cont = saved;
                
                env.setAll(fVars,vs);
                body.accept(Eval_Exp.this);
                throw new Error(); // can never get here...
            }
        });
        
        exps.accept(eval_explist);
        throw new Error(); // can never get here...
    }

    private void applyPrim (byte op, List args) throws VisitE {
        int arity = Prim.opArity(op);
        SchemeValue arg1 = null;
        SchemeValue arg2 = null;
        if (arity == 1) 
            arg1 = (SchemeValue) args.hd();
        else if (arity == 2) { 
            arg1 = (SchemeValue) args.hd();
            arg2 = (SchemeValue) args.tl().hd();
        }
        else ;
        int v1 = 0;
        int v2 = 0;

        switch (op) {
        case Prim.MINUS: 
            v1 = arg1.getInt();
            v2 = arg2.getInt();
            cont.invoke(new IntV(v1 - v2)); break;
        case Prim.TIMES: 
            v1 = arg1.getInt();
            v2 = arg2.getInt();
            cont.invoke(new IntV(v1 * v2)); break;
        case Prim.PLUS: 
            v1 = arg1.getInt();
            v2 = arg2.getInt();
            cont.invoke(new IntV(v1 + v2)); break;
        case Prim.LESS: 
            v1 = arg1.getInt();
            v2 = arg2.getInt();
            cont.invoke(new BoolV(v1 < v2)); break;
        case Prim.LEQ: 
            v1 = arg1.getInt();
            v2 = arg2.getInt();
            cont.invoke(new BoolV(v1 <= v2)); break;
        case Prim.EQ: 
            v1 = arg1.getInt();
            v2 = arg2.getInt();
            cont.invoke(new BoolV(v1 == v2)); break;
        case Prim.EQP: 
            cont.invoke(new BoolV(arg1 == arg2)); break;
        case Prim.CONS: 
            cont.invoke(new PairV(arg1,arg2)); break;
        case Prim.SETCAR: 
            arg1.setCar(arg2);
            cont.invoke(new VoidV()); break;
        case Prim.SETCDR: 
            arg1.setCdr(arg2);
            cont.invoke(new VoidV()); break;
        case Prim.CAR: 
            cont.invoke(arg1.getCar()); break;
        case Prim.CDR: 
            cont.invoke(arg1.getCdr()); break;
        case Prim.ADD1: 
            v1 = arg1.getInt();
            cont.invoke(new IntV(v1+1)); break;
        case Prim.SUB1: 
            v1 = arg1.getInt();
            cont.invoke(new IntV(v1-1)); break;
        case Prim.NOT: 
            if (arg1.isTrue()) cont.invoke(new BoolV(false)); 
            else cont.invoke(new BoolV(true)); 
            break;
        case Prim.BOOLEANP: 
            if (arg1 instanceof BoolV) cont.invoke(new BoolV(true)); 
            else cont.invoke(new BoolV(false)); 
            break;
        case Prim.INTEGERP: 
            if (arg1 instanceof IntV) cont.invoke(new BoolV(true)); 
            else cont.invoke(new BoolV(false)); 
            break;
        case Prim.NULLP: 
            if (arg1 instanceof EmptyV) cont.invoke(new BoolV(true)); 
            else cont.invoke(new BoolV(false)); 
            break;
        case Prim.PAIRP: 
            if (arg1 instanceof PairV) cont.invoke(new BoolV(true)); 
            else cont.invoke(new BoolV(false)); 
            break;
        case Prim.PROCEDUREP: 
            if (arg1 instanceof ClosureV) cont.invoke(new BoolV(true)); 
            else cont.invoke(new BoolV(false)); 
            break;
        case Prim.ZEROP: 
            v1 = arg1.getInt();
            if (v1 == 0) cont.invoke(new BoolV(true)); 
            else cont.invoke(new BoolV(false)); 
            break;
        case Prim.CALLCC:
            List ccArgs = new Cons(cont, new EmptyL());
            arg1.apply(ccArgs,this);
            throw new Error(); // can never get here...
        case Prim.VOID: 
            throw new VisitE("Cannot apply primitive void");
        default: throw new VisitE("Unknown operator");
        }
    }
}
